/*

Program Title:          Mega Pi

Source File(s):         megapi.c

Last Modified:          2000-02-16

Author:                 Alan Pittman <ahpittman@yahoo.com>

Compiler:               Watcom C/C++ 10.6

Program Type:           32-bit Character-Mode Executable
                        running in a MS-DOS Window

Purpose:                Calculate Pi up to 1 million digits
                        using the Gauss Arctangent Algorithm

References:

Usage:  megapi <number_of_digits>

-----------------------------------------------------------------------------*/

/*
The x87 series coprocessors have a "round to integer" instruction, which is
exceedingly slow. The same function can be achieved very much faster by adding
a very large magic number in the form 3*2^n to the result and then subtracting
it. This causes the fractional bits to be shifted out.

This must be done in such a way that does not cause the optimizer to simply
remove both add and subtract operations because it considers them redundant.

The x87 rounding mode must be set to truncation so that the fractional bits
are simply discarded.

This program dynamically determines what the magic number is.
*/

#define USEMAGIC 1      // If Magic Number Trick fails, recompile with 0

#include <float.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define RADIX 1000000000.0
#define EMPTY 999999999

struct bignum           // structure for a Big Number
{
    long size;          // number of doubles
    double *array;      // each double contains 9 decimals
    long start;         // postion of first non-zero double
};

void bigNew(struct bignum *x, long size, long value);
void bigShow(struct bignum *x, long digits);

void bigDivMov(struct bignum *y, struct bignum *x, long n);
void bigDivAdd(struct bignum *y, struct bignum *x, long n);
void bigDivSub(struct bignum *y, struct bignum *x, long n);

void bigCarry(struct bignum *x);
void bigOverflow(struct bignum *x);
void bigUnderflow(struct bignum *x);

// To fool the optimizer,
// keep two independant versions of the Magic Number.

#if USEMAGIC == 1
double MAGIC;
double MAGIK;

unsigned short int SetFPUMode(unsigned short int Mode,unsigned short int Mask)
{static short unsigned int CtrlWord;

asm volatile ("
        fstcw   (%0)
        fwait
       "
        : :"m" (CtrlWord) : "cc" , "memory");

CtrlWord= (CtrlWord & ~ Mask) | (Mode & Mask);
CtrlWord=0xffff;

asm volatile ("
        fldcw   (%0)
        fwait
       "
        : :"m" (CtrlWord) : "cc", "memory");

return CtrlWord;
}
#endif

//----------------------------------------------------------------
//  mainline
//----------------------------------------------------------------

void main(int argc, char *argv[])
{
    struct bignum a, b, c, d, pi;
    long digits, size, n, s;
    clock_t start, end;

#if USEMAGIC == 1
    unsigned int status;
    double test87;
    long power = 51;
#endif

    // Get the Number of Digits required

    if(argc != 2) {
        printf("\nUsage: megapi number_of_digits\n\n");
        system("PAUSE");
        exit(1);
    }
    digits = atol(argv[1]);

#if USEMAGIC == 1

    // Determine the correct Magic Number.
    // This tricks the processor into removing the fractional part of a
    // floating point number thus converting it to an integer value.
    // This is very much faster than the following:
    //          double_variable = (long) double_variable; 

/*    status = _control87(0xC00, 0xC00);*/
    status = SetFPUMode(0xC00, 0xC00);
    MAGIC = 6755399441055744.0;         // start with 3 * 2 ^ 51
    
    MAGIK = 3.0;                       // get the same value
    while (MAGIK < 6.0e15)             // by devious means
        MAGIK *= 2.0;                  // to out-fox the optimizer

    for (power = 51;; power++) {
        test87 = 999999999.999999;      // this value creates most stress
        test87 += MAGIC;                // force the loss of low-end bits
        test87 -= MAGIK;               // reverse the add with a subtract
        if (test87 == 999999999.0)      // float converted to integer?
            break;                      // bingo!
            
        if (power == 70) {              // we are not having any luck
            printf("floating-point problem\n");
            printf("recompile program with:\n");
            printf("    #define USEMAGIC 0\n");
            system("PAUSE");
            exit(1);
         }
         
        MAGIC *= 2.0;                   // double the magic number(s)
        MAGIK *= 2.0;                   // and try again
    }
    
    printf("Magic Number = 3 * 2 ^ %d\n", power);
#endif

    printf("Number of digits = %d\n", digits);

    start = clock();                            // start timing

    size = 1 + (digits + 8) / 9 + 2;            // add some extra digits

    bigNew(&a,  size, 48);                      //  a = 48;
    bigNew(&b,  size, 32);                      //  b = 32;
    bigNew(&c,  size, 20);                      //  c = 20;
    bigNew(&d,  size,  0);
    bigNew(&pi, size,  0);

    bigDivMov(&pi, &a, 18);                     //  a = a / 18;  pi = a;
    bigDivAdd(&pi, &b, 57);                     //  b = b / 57;  pi = pi + b;
    bigDivSub(&pi, &c, 239);                    //  c = c / 239; pi = pi - c;

    bigCarry(&pi);

    for (n = 3, s = -1;; n += 2, s = -s) {
        bigDivMov(&d, &a, 324);                 //  a = a / 324;   d = a; 
        if (a.start == EMPTY)
            break;                              //  a is zero, finished! 
        if (b.start != EMPTY) {
            bigDivAdd(&d, &b, 3249);            //  b = b / 3249;  d = d + b;
            if (c.start != EMPTY) {
                bigDivSub(&d, &c, 57121);       //  c = c / 57121; d = d - b;
                bigCarry(&d);
            } else {
                bigOverflow(&d);
            }
        }
        if (s == -1) {
            bigDivSub(&pi, &d, n);              //  d = d / n; pi = pi - d;
            bigUnderflow(&pi);
        } else {
            bigDivAdd(&pi, &d, n);              //  d = d / n; pi = pi + d;
            bigOverflow(&pi);
        }
        if (n % 800 == 1) {
            end = clock();
            printf("n = %d    time = %0.2f\n",
                    ((n - 1)*5)/4, (float) (end - start) / CLOCKS_PER_SEC);
        }
    }

    end = clock();              // don't include time to write file

    bigShow(&pi,digits);        // save the result

    free(a.array);              // free the large numbers
    free(b.array);
    free(c.array);
    free(d.array);
    free(pi.array);

    printf("%d digits of Pi are saved in file: pi.dat\n", digits);
    printf("time = %0.2f sec\n", (float) (end - start) / CLOCKS_PER_SEC);
}

//----------------------------------------------------------------
//  bigNew - Allocate and initialize a big number
//----------------------------------------------------------------

void bigNew(struct bignum *x, long size, long value)
{
    long i;

    x->size = size;
    x->array = (double *) malloc(size * sizeof(double));
    if (x->array == NULL) {
        printf("malloc error\n");
        system("PAUSE");
        exit(1);
    }
    x->array[0] = (double) value;
    for (i = 1; i < x->size; i++)
        x->array[i] = 0.0;
    if (value == 0)
        x->start = EMPTY;
    else
        x->start = 0;
}

//----------------------------------------------------------------
//  bigShow - Display and save result 
//----------------------------------------------------------------

void bigShow(struct bignum *x, long digits)
{
    long i, j, k;
    char s[20];
    FILE *fp;

    fp = fopen("pi.dat", "w");
    printf("%d.", (long) x->array[0]);
    fprintf(fp, "%d.", (long) x->array[0]);
    for (i = 1, k = 0; i < x->size; i++) {
        sprintf(s, "%09d", (long) x->array[i]);
        for (j = 0; j < 9 && k < digits; j++, k++) {
            if (k == 0) {
                printf("%c", s[j]);
                fprintf(fp,"%c", s[j]);
            }
            else if (k % 1000 == 999) {
                printf("%c  :%d", s[j],k+1);
                fprintf(fp,"%c  :%d", s[j],k+1);
            }
            else if (k % 1000 == 0) {
                printf("\n\n  %c", s[j]);
                fprintf(fp,"\n\n  %c", s[j]);
            }
            else if (k % 50 == 0) {
                printf("\n  %c", s[j]);
                fprintf(fp,"\n  %c", s[j]);
            }
            else if (k % 5 == 0) {
                printf(" %c", s[j]);
                fprintf(fp," %c", s[j]);
            }
            else {
                printf("%c", s[j]);
                fprintf(fp,"%c", s[j]);
            }
        }
    }
    printf("\n\n");
    fprintf(fp,"\n\n");

    fclose(fp);
}

//----------------------------------------------------------------
//  bigDivSub - Divide and Move
//
//              X = X / n
//              Y = X
//----------------------------------------------------------------

void bigDivMov(struct bignum *y, struct bignum *x, long n)
{
    long i, newStart = EMPTY;
    double quotient, remainder = 0.0;
    double dividend, denominator, reciprocal;

#if USEMAGIC == 1
    static double magic, magik;
    magic = MAGIC;
    magik = MAGIK;
#endif

    denominator = (double) n;
    reciprocal = 1.0 / denominator;

    for (i = y->start; i < x->start; i++) {
        y->array[i] = 0.0;
    }

    for (i = x->start; i < x->size; i++) {
        dividend = remainder * RADIX + x->array[i];
#if USEMAGIC == 1
        quotient = dividend * reciprocal;               // fast method
        quotient += magic;
        quotient -= magik;
#else
        quotient = (long) (dividend * reciprocal);      // slow method
#endif
        remainder = dividend - denominator * quotient;
        x->array[i] = quotient;
        y->array[i] = quotient;
        if (newStart == EMPTY && quotient > 0.0)
            newStart = i;
    }
    x->start = newStart;
    y->start = newStart;
}

//----------------------------------------------------------------
//  bigDivAdd - Divide and Add
//
//              X = X / n
//              Y = Y + X
//----------------------------------------------------------------

void bigDivAdd(struct bignum *y, struct bignum *x, long n)
{
    long i, newStart = EMPTY;
    double quotient, remainder = 0.0;
    double dividend, denominator, reciprocal;

#if USEMAGIC == 1
    static double magic, magik;
    magic = MAGIC;
    magik = MAGIK;
#endif

    denominator = (double) n;
    reciprocal = 1.0 / denominator;

    for (i = x->start; i < x->size; i++) {
        dividend = remainder * RADIX + x->array[i];
#if USEMAGIC == 1
        quotient = dividend * reciprocal;               // fast method
        quotient += magic;
        quotient -= magik;
#else
        quotient = (long) (dividend * reciprocal);      // slow method
#endif
        remainder = dividend - denominator * quotient;
        x->array[i] = quotient;
        y->array[i] += quotient;
        if (newStart == EMPTY && quotient > 0.0)
            newStart = i;
    }
    x->start = newStart;
}

//----------------------------------------------------------------
//  bigDivSub - Divide and Subtract
//
//              X = X / n
//              Y = Y - X
//----------------------------------------------------------------

void bigDivSub(struct bignum *y, struct bignum *x, long n)
{
    long i, newStart = EMPTY;
    double quotient, remainder = 0.0;
    double dividend, denominator, reciprocal;

#if USEMAGIC == 1
    static double magic, magik;
    magic = MAGIC;
    magik = MAGIK;
#endif

    denominator = (double) n;
    reciprocal = 1.0 / denominator;

    for (i = x->start; i < x->size; i++) {
        dividend = remainder * RADIX + x->array[i];
#if USEMAGIC == 1
        quotient = dividend * reciprocal;               // fast method
        quotient += magic;
        quotient -= magik;
#else
        quotient = (long) (dividend * reciprocal);      // slow method
#endif
        remainder = dividend - denominator * quotient;
        x->array[i] = quotient;
        y->array[i] -= quotient;
        if (newStart == EMPTY && quotient > 0.0)
            newStart = i;
    }
    x->start = newStart;
}

//----------------------------------------------------------------
//  bigCarry - Release both underflow and overflow carries
//----------------------------------------------------------------

void bigCarry(struct bignum *x)
{
    long i, newStart = EMPTY;
    double sum, carry = 0.0;

    for (i = x->size - 1;; i--) {
        sum = x->array[i] + carry;
        if (sum < 0.0) {
            sum += RADIX;
            carry = -1.0;
        } else {
            if (sum >= RADIX) {
                sum -= RADIX;
                carry = +1.0;
            } else {
                carry = 0.0;
            }
        }
        x->array[i] = sum;
        if (sum > 0.0)
            newStart = i;
        if (carry == 0.0 && i <= x->start)
            break;
    }
    x->start = newStart;
}

//----------------------------------------------------------------
//  bigOverflow - Release overflow carries
//----------------------------------------------------------------

void bigOverflow(struct bignum *x)
{
    long i, newStart = EMPTY;
    double sum, carry = 0.0;

    for (i = x->size - 1;; i--) {
        sum = x->array[i] + carry;
        if (sum >= RADIX) {
            sum -= RADIX;
            carry = +1.0;
        } else {
            carry = 0.0;
        }
        x->array[i] = sum;
        if (sum > 0.0)
            newStart = i;
        if (carry == 0.0 && i <= x->start)
            break;
    }
    x->start = newStart;
}

//----------------------------------------------------------------
//  bigUnderflow - Release underflow carries
//----------------------------------------------------------------

void bigUnderflow(struct bignum *x)
{
    long i, newStart = EMPTY;
    double sum, carry = 0.0;

    for (i = x->size - 1;; i--) {
        sum = x->array[i] + carry;
        if (sum < 0.0) {
            sum += RADIX;
            carry = -1.0;
        } else {
            carry = 0.0;
        }
        x->array[i] = sum;
        if (sum > 0.0)
            newStart = i;
        if (carry == 0.0 && i <= x->start)
            break;
    }
    x->start = newStart;
}

// End of Megapi.c
